@@ -1,7 +1,6 @@
-/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
- * vim: set ts=8 sw=4 et tw=78:
- *
- * This Source Code Form is subject to the terms of the Mozilla Public
+/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* vim: set ts=8 sts=4 et sw=4 tw=99: */
+/* This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
@@ -67,7 +66,7 @@ UnwrapNativeCPOW(nsISupports* wrapper)
     nsCOMPtr<nsIXPConnectWrappedJS> underware = do_QueryInterface(wrapper);
     if (underware) {
         JSObject* mainObj = underware->GetJSObject();
-        if (mainObj && mozilla::jsipc::JavaScriptParent::IsCPOW(mainObj))
+        if (mainObj && mozilla::jsipc::IsCPOW(mainObj))
             return mainObj;
     }
     return nullptr;
@@ -105,242 +104,245 @@ XPCConvert::NativeData2JS(MutableHandleValue d, const void* s,
         *pErr = NS_ERROR_XPC_BAD_CONVERT_NATIVE;
 
     switch (type.TagPart()) {
-    case nsXPTType::T_I8    : d.setInt32(*((int8_t*)s));             break;
-    case nsXPTType::T_I16   : d.setInt32(*((int16_t*)s));            break;
-    case nsXPTType::T_I32   : d.setInt32(*((int32_t*)s));            break;
-    case nsXPTType::T_I64   : d.setNumber(double(*((int64_t*)s)));   break;
-    case nsXPTType::T_U8    : d.setInt32(*((uint8_t*)s));            break;
-    case nsXPTType::T_U16   : d.setInt32(*((uint16_t*)s));           break;
-    case nsXPTType::T_U32   : d.setNumber(*((uint32_t*)s));          break;
-    case nsXPTType::T_U64   : d.setNumber(double(*((uint64_t*)s)));  break;
-    case nsXPTType::T_FLOAT : d.setNumber(*((float*)s));             break;
-    case nsXPTType::T_DOUBLE: d.setNumber(*((double*)s));            break;
+    case nsXPTType::T_I8    :
+        d.setInt32(*static_cast<const int8_t*>(s));
+        return true;
+    case nsXPTType::T_I16   :
+        d.setInt32(*static_cast<const int16_t*>(s));
+        return true;
+    case nsXPTType::T_I32   :
+        d.setInt32(*static_cast<const int32_t*>(s));
+        return true;
+    case nsXPTType::T_I64   :
+        d.setNumber(static_cast<double>(*static_cast<const int64_t*>(s)));
+        return true;
+    case nsXPTType::T_U8    :
+        d.setInt32(*static_cast<const uint8_t*>(s));
+        return true;
+    case nsXPTType::T_U16   :
+        d.setInt32(*static_cast<const uint16_t*>(s));
+        return true;
+    case nsXPTType::T_U32   :
+        d.setNumber(*static_cast<const uint32_t*>(s));
+        return true;
+    case nsXPTType::T_U64   :
+        d.setNumber(static_cast<double>(*static_cast<const uint64_t*>(s)));
+        return true;
+    case nsXPTType::T_FLOAT :
+        d.setNumber(*static_cast<const float*>(s));
+        return true;
+    case nsXPTType::T_DOUBLE:
+        d.setNumber(*static_cast<const double*>(s));
+        return true;
     case nsXPTType::T_BOOL  :
-        {
-            bool b = *((bool*)s);
-
-            NS_WARN_IF_FALSE(b == 1 || b == 0,
-                             "Passing a malformed bool through XPConnect");
-            d.setBoolean(b);
-            break;
-        }
+        d.setBoolean(*static_cast<const bool*>(s));
+        return true;
     case nsXPTType::T_CHAR  :
-        {
-            char* p = (char*)s;
-            if (!p)
-                return false;
+    {
+        char p = *static_cast<const char*>(s);
 
 #ifdef STRICT_CHECK_OF_UNICODE
-            MOZ_ASSERT(! ILLEGAL_CHAR_RANGE(p) , "passing non ASCII data");
+        MOZ_ASSERT(! ILLEGAL_CHAR_RANGE(p) , "passing non ASCII data");
 #endif // STRICT_CHECK_OF_UNICODE
 
-            JSString* str;
-            if (!(str = JS_NewStringCopyN(cx, p, 1)))
-                return false;
-            d.setString(str);
-            break;
-        }
+        JSString* str = JS_NewStringCopyN(cx, &p, 1);
+        if (!str)
+            return false;
+
+        d.setString(str);
+        return true;
+    }
     case nsXPTType::T_WCHAR :
-        {
-            jschar* p = (jschar*)s;
-            if (!p)
-                return false;
-            JSString* str;
-            if (!(str = JS_NewUCStringCopyN(cx, p, 1)))
-                return false;
-            d.setString(str);
-            break;
-        }
+    {
+        jschar p = *static_cast<const jschar*>(s);
+
+        JSString* str = JS_NewUCStringCopyN(cx, &p, 1);
+        if (!str)
+            return false;
+
+        d.setString(str);
+        return true;
+    }
 
     case nsXPTType::T_JSVAL :
-        {
-            d.set(*((Value*)s));
-            if (!JS_WrapValue(cx, d))
-                return false;
-            break;
+    {
+        d.set(*static_cast<const Value*>(s));
+        return JS_WrapValue(cx, d);
+    }
+
+    case nsXPTType::T_VOID:
+        XPC_LOG_ERROR(("XPCConvert::NativeData2JS : void* params not supported"));
+        return false;
+
+    case nsXPTType::T_IID:
+    {
+        nsID* iid2 = *static_cast<nsID* const *>(s);
+        if (!iid2) {
+            d.setNull();
+            return true;
         }
 
-    default:
-
-        // set the default result
-        d.setNull();
-
-        switch (type.TagPart()) {
-        case nsXPTType::T_VOID:
-            XPC_LOG_ERROR(("XPCConvert::NativeData2JS : void* params not supported"));
+        RootedObject scope(cx, JS::CurrentGlobalOrNull(cx));
+        JSObject* obj = xpc_NewIDObject(cx, scope, *iid2);
+        if (!obj)
             return false;
 
-        case nsXPTType::T_IID:
-            {
-                nsID* iid2 = *((nsID**)s);
-                if (!iid2)
-                    break;
-                RootedObject scope(cx, JS::CurrentGlobalOrNull(cx));
-                JSObject* obj;
-                if (!(obj = xpc_NewIDObject(cx, scope, *iid2)))
-                    return false;
-                d.setObject(*obj);
-                break;
-            }
+        d.setObject(*obj);
+        return true;
+    }
 
-        case nsXPTType::T_ASTRING:
-            // Fall through to T_DOMSTRING case
+    case nsXPTType::T_ASTRING:
+        // Fall through to T_DOMSTRING case
 
-        case nsXPTType::T_DOMSTRING:
-            {
-                const nsAString* p = *((const nsAString**)s);
-                if (!p)
-                    break;
+    case nsXPTType::T_DOMSTRING:
+    {
+        const nsAString* p = *static_cast<const nsAString* const *>(s);
+        if (!p || p->IsVoid()) {
+            d.setNull();
+            return true;
+        }
 
-                if (!p->IsVoid()) {
-                    nsStringBuffer* buf;
-                    if (!XPCStringConvert::ReadableToJSVal(cx, *p, &buf, d))
-                        return false;
-                    if (buf)
-                        buf->AddRef();
-                }
+        nsStringBuffer* buf;
+        if (!XPCStringConvert::ReadableToJSVal(cx, *p, &buf, d))
+            return false;
+        if (buf)
+            buf->AddRef();
+        return true;
+    }
 
-                // *d is defaulted to JSVAL_NULL so no need to set it
-                // again if p is a "void" string
-                MOZ_ASSERT_IF(p->IsVoid(), d.isNull());
-                break;
-            }
-
-        case nsXPTType::T_CHAR_STR:
-            {
-                char* p = *((char**)s);
-                if (!p)
-                    break;
+    case nsXPTType::T_CHAR_STR:
+    {
+        const char* p = *static_cast<const char* const *>(s);
+        if (!p) {
+            d.setNull();
+            return true;
+        }
 
 #ifdef STRICT_CHECK_OF_UNICODE
-                bool isAscii = true;
-                char* t;
-                for (t=p; *t && isAscii ; t++) {
-                  if (ILLEGAL_CHAR_RANGE(*t))
-                      isAscii = false;
-                }
-                MOZ_ASSERT(isAscii, "passing non ASCII data");
+        bool isAscii = true;
+        for (char* t = p; *t && isAscii; t++) {
+          if (ILLEGAL_CHAR_RANGE(*t))
+              isAscii = false;
+        }
+        MOZ_ASSERT(isAscii, "passing non ASCII data");
 #endif // STRICT_CHECK_OF_UNICODE
-                JSString* str;
-                if (!(str = JS_NewStringCopyZ(cx, p)))
-                    return false;
-                d.setString(str);
-                break;
-            }
 
-        case nsXPTType::T_WCHAR_STR:
-            {
-                jschar* p = *((jschar**)s);
-                if (!p)
-                    break;
-                JSString* str;
-                if (!(str = JS_NewUCStringCopyZ(cx, p)))
-                    return false;
-                d.setString(str);
-                break;
-            }
-        case nsXPTType::T_UTF8STRING:
-            {
-                const nsACString* utf8String = *((const nsACString**)s);
+        JSString* str = JS_NewStringCopyZ(cx, p);
+        if (!str)
+            return false;
 
-                if (!utf8String || utf8String->IsVoid())
-                    break;
+        d.setString(str);
+        return true;
+    }
 
-                if (utf8String->IsEmpty()) {
-                    d.set(JS_GetEmptyStringValue(cx));
-                    break;
-                }
+    case nsXPTType::T_WCHAR_STR:
+    {
+        const jschar* p = *static_cast<const jschar* const *>(s);
+        if (!p) {
+            d.setNull();
+            return true;
+        }
 
-                const uint32_t len = CalcUTF8ToUnicodeLength(*utf8String);
-                // The cString is not empty at this point, but the calculated
-                // UTF-16 length is zero, meaning no valid conversion exists.
-                if (!len)
-                    return false;
+        JSString* str = JS_NewUCStringCopyZ(cx, p);
+        if (!str)
+            return false;
 
-                const size_t buffer_size = (len + 1) * sizeof(char16_t);
-                char16_t* buffer =
-                    static_cast<char16_t*>(JS_malloc(cx, buffer_size));
-                if (!buffer)
-                    return false;
+        d.setString(str);
+        return true;
+    }
+    case nsXPTType::T_UTF8STRING:
+    {
+        const nsACString* utf8String = *static_cast<const nsACString* const *>(s);
 
-                uint32_t copied;
-                if (!UTF8ToUnicodeBuffer(*utf8String, buffer, &copied) ||
-                    len != copied) {
-                    // Copy or conversion during copy failed. Did not copy the
-                    // whole string.
-                    JS_free(cx, buffer);
-                    return false;
-                }
+        if (!utf8String || utf8String->IsVoid()) {
+            d.setNull();
+            return true;
+        }
 
-                // JS_NewUCString takes ownership on success, i.e. a
-                // successful call will make it the responsiblity of the JS VM
-                // to free the buffer.
-                JSString* str = JS_NewUCString(cx, (jschar*)buffer, len);
-                if (!str) {
-                    JS_free(cx, buffer);
-                    return false;
-                }
+        if (utf8String->IsEmpty()) {
+            d.set(JS_GetEmptyStringValue(cx));
+            return true;
+        }
 
-                d.setString(str);
-                break;
-            }
-        case nsXPTType::T_CSTRING:
-            {
-                const nsACString* cString = *((const nsACString**)s);
+        const uint32_t len = CalcUTF8ToUnicodeLength(*utf8String);
+        // The cString is not empty at this point, but the calculated
+        // UTF-16 length is zero, meaning no valid conversion exists.
+        if (!len)
+            return false;
 
-                if (!cString || cString->IsVoid())
-                    break;
+        const size_t buffer_size = (len + 1) * sizeof(char16_t);
+        char16_t* buffer =
+            static_cast<char16_t*>(JS_malloc(cx, buffer_size));
+        if (!buffer)
+            return false;
 
-                if (cString->IsEmpty()) {
-                    d.set(JS_GetEmptyStringValue(cx));
-                    break;
-                }
-
-                // c-strings (binary blobs) are deliberately not converted from
-                // UTF-8 to UTF-16. T_UTF8Sting is for UTF-8 encoded strings
-                // with automatic conversion.
-                JSString* str = JS_NewStringCopyN(cx, cString->Data(),
-                                                  cString->Length());
-                if (!str)
-                    return false;
-
-                d.setString(str);
-                break;
-            }
-
-        case nsXPTType::T_INTERFACE:
-        case nsXPTType::T_INTERFACE_IS:
-            {
-                nsISupports* iface = *((nsISupports**)s);
-                if (iface) {
-                    if (iid->Equals(NS_GET_IID(nsIVariant))) {
-                        nsCOMPtr<nsIVariant> variant = do_QueryInterface(iface);
-                        if (!variant)
-                            return false;
-
-                        return XPCVariant::VariantDataToJS(variant,
-                                                           pErr, d);
-                    }
-                    // else...
-                    xpcObjectHelper helper(iface);
-                    if (!NativeInterface2JSObject(d, nullptr, helper, iid,
-                                                  nullptr, true, pErr))
-                        return false;
-
-#ifdef DEBUG
-                    JSObject* jsobj = d.toObjectOrNull();
-                    if (jsobj && !js::GetObjectParent(jsobj))
-                        MOZ_ASSERT(js::GetObjectClass(jsobj)->flags & JSCLASS_IS_GLOBAL,
-                                   "Why did we recreate this wrapper?");
-#endif
-                }
-                break;
-            }
-
-        default:
-            NS_ERROR("bad type");
+        uint32_t copied;
+        if (!UTF8ToUnicodeBuffer(*utf8String, buffer, &copied) ||
+            len != copied) {
+            // Copy or conversion during copy failed. Did not copy the
+            // whole string.
+            JS_free(cx, buffer);
             return false;
         }
+
+        // JS_NewUCString takes ownership on success, i.e. a
+        // successful call will make it the responsiblity of the JS VM
+        // to free the buffer.
+        JSString* str = JS_NewUCString(cx, buffer, len);
+        if (!str) {
+            JS_free(cx, buffer);
+            return false;
+        }
+
+        d.setString(str);
+        return true;
+    }
+    case nsXPTType::T_CSTRING:
+    {
+        const nsACString* cString = *static_cast<const nsACString* const *>(s);
+
+        if (!cString || cString->IsVoid()) {
+            d.setNull();
+            return true;
+        }
+
+        // c-strings (binary blobs) are deliberately not converted from
+        // UTF-8 to UTF-16. T_UTF8Sting is for UTF-8 encoded strings
+        // with automatic conversion.
+        JSString* str = JS_NewStringCopyN(cx, cString->Data(),
+                                          cString->Length());
+        if (!str)
+            return false;
+
+        d.setString(str);
+        return true;
+    }
+
+    case nsXPTType::T_INTERFACE:
+    case nsXPTType::T_INTERFACE_IS:
+    {
+        nsISupports* iface = *static_cast<nsISupports* const *>(s);
+        if (!iface) {
+            d.setNull();
+            return true;
+        }
+
+        if (iid->Equals(NS_GET_IID(nsIVariant))) {
+            nsCOMPtr<nsIVariant> variant = do_QueryInterface(iface);
+            if (!variant)
+                return false;
+
+            return XPCVariant::VariantDataToJS(variant,
+                                               pErr, d);
+        }
+
+        xpcObjectHelper helper(iface);
+        return NativeInterface2JSObject(d, nullptr, helper, iid, nullptr, true, pErr);
+    }
+
+    default:
+        NS_ERROR("bad type");
+        return false;
     }
     return true;
 }
@@ -379,8 +381,6 @@ XPCConvert::JSData2Native(void* d, HandleValue s,
 {
     NS_PRECONDITION(d, "bad param");
 
-    bool isDOMString = true;
-
     AutoJSContext cx;
     if (pErr)
         *pErr = NS_ERROR_XPC_BAD_CONVERT_JS;
@@ -473,88 +473,77 @@ XPCConvert::JSData2Native(void* d, HandleValue s,
 
     case nsXPTType::T_ASTRING:
     {
-        isDOMString = false;
+        if (s.isUndefined()) {
+            if (useAllocator)
+                *((const nsAString**)d) = &NullString();
+            else
+                (**((nsAString**)d)).SetIsVoid(true);
+            return true;
+        }
         // Fall through to T_DOMSTRING case.
     }
     case nsXPTType::T_DOMSTRING:
     {
-        static const char16_t EMPTY_STRING[] = { '\0' };
-        static const char16_t VOID_STRING[] = { 'u', 'n', 'd', 'e', 'f', 'i', 'n', 'e', 'd', '\0' };
-
+        if (s.isNull()) {
+            if (useAllocator)
+                *((const nsAString**)d) = &NullString();
+            else
+                (**((nsAString**)d)).SetIsVoid(true);
+            return true;
+        }
+        size_t length = 0;
         const char16_t* chars = nullptr;
         JSString* str = nullptr;
-        bool isNewString = false;
-        uint32_t length = 0;
-
-        if (JSVAL_IS_VOID(s)) {
-            if (isDOMString) {
-                chars  = VOID_STRING;
-                length = ArrayLength(VOID_STRING) - 1;
-            } else {
-                chars = EMPTY_STRING;
-                length = 0;
-            }
-        } else if (!JSVAL_IS_NULL(s)) {
+        if (!s.isUndefined()) {
             str = ToString(cx, s);
             if (!str)
                 return false;
 
-            length = (uint32_t) JS_GetStringLength(str);
-            if (length) {
-                chars = JS_GetStringCharsZ(cx, str);
-                if (!chars)
-                    return false;
-                if (STRING_TO_JSVAL(str) != s)
-                    isNewString = true;
-            } else {
-                str = nullptr;
-                chars = EMPTY_STRING;
+            chars = useAllocator ? JS_GetStringCharsZAndLength(cx, str, &length)
+                                 : JS_GetStringCharsAndLength(cx, str, &length);
+            if (!chars)
+                return false;
+
+            if (!length) {
+                if (useAllocator)
+                    *((const nsAString**)d) = &EmptyString();
+                else
+                    (**((nsAString**)d)).Truncate();
+                return true;
             }
         }
 
+        nsString* ws;
         if (useAllocator) {
-            // XXX extra string copy when isNewString
-            if (str && !isNewString) {
-                size_t strLength;
-                const jschar *strChars = JS_GetStringCharsZAndLength(cx, str, &strLength);
-                if (!strChars)
-                    return false;
-
-                XPCReadableJSStringWrapper *wrapper =
-                    nsXPConnect::GetRuntimeInstance()->NewStringWrapper(strChars, strLength);
-                if (!wrapper)
-                    return false;
-
-                *((const nsAString**)d) = wrapper;
-            } else if (JSVAL_IS_NULL(s)) {
-                XPCReadableJSStringWrapper *wrapper =
-                    new XPCReadableJSStringWrapper();
-                if (!wrapper)
-                    return false;
-
-                *((const nsAString**)d) = wrapper;
-            } else {
-                // use nsString to encourage sharing
-                const nsAString *rs = new nsString(chars, length);
-                if (!rs)
-                    return false;
-                *((const nsAString**)d) = rs;
-            }
+            ws = nsXPConnect::GetRuntimeInstance()->NewShortLivedString();
+            *((const nsString**)d) = ws;
         } else {
-            nsAString* ws = *((nsAString**)d);
+            ws = *((nsString**)d);
+        }
 
-            if (JSVAL_IS_NULL(s) || (!isDOMString && JSVAL_IS_VOID(s))) {
-                ws->Truncate();
-                ws->SetIsVoid(true);
-            } else
-                ws->Assign(chars, length);
+        if (!str) {
+            ws->AssignLiteral(MOZ_UTF16("undefined"));
+        } else if (XPCStringConvert::IsDOMString(str)) {
+            // The characters represent an existing nsStringBuffer that
+            // was shared by XPCStringConvert::ReadableToJSVal.
+            nsStringBuffer::FromData((void *)chars)->ToString(length, *ws);
+        } else if (XPCStringConvert::IsLiteral(str)) {
+            // The characters represent a literal char16_t string constant
+            // compiled into libxul, such as the string "undefined" above.
+            ws->AssignLiteral(chars, length);
+        } else if (useAllocator && STRING_TO_JSVAL(str) == s) {
+            // The JS string will exist over the function call.
+            // We don't need to copy the characters in this case.
+            ws->Rebind(chars, length);
+        } else {
+            ws->Assign(chars, length);
         }
         return true;
     }
 
     case nsXPTType::T_CHAR_STR:
     {
-        if (JSVAL_IS_VOID(s) || JSVAL_IS_NULL(s)) {
+        if (s.isUndefined() || s.isNull()) {
             *((char**)d) = nullptr;
             return true;
         }
@@ -595,7 +584,7 @@ XPCConvert::JSData2Native(void* d, HandleValue s,
         const jschar* chars=nullptr;
         JSString* str;
 
-        if (JSVAL_IS_VOID(s) || JSVAL_IS_NULL(s)) {
+        if (s.isUndefined() || s.isNull()) {
             *((jschar**)d) = nullptr;
             return true;
         }
@@ -622,20 +611,14 @@ XPCConvert::JSData2Native(void* d, HandleValue s,
     case nsXPTType::T_UTF8STRING:
     {
         const jschar* chars;
-        uint32_t length;
+        size_t length;
         JSString* str;
 
-        if (JSVAL_IS_NULL(s) || JSVAL_IS_VOID(s)) {
+        if (s.isNull() || s.isUndefined()) {
             if (useAllocator) {
-                nsACString *rs = new nsCString();
-                if (!rs)
-                    return false;
-
-                rs->SetIsVoid(true);
-                *((nsACString**)d) = rs;
+                *((const nsACString**)d) = &NullCString();
             } else {
                 nsCString* rs = *((nsCString**)d);
-                rs->Truncate();
                 rs->SetIsVoid(true);
             }
             return true;
@@ -644,11 +627,19 @@ XPCConvert::JSData2Native(void* d, HandleValue s,
         // The JS val is neither null nor void...
 
         if (!(str = ToString(cx, s))||
-            !(chars = JS_GetStringCharsZ(cx, str))) {
+            !(chars = JS_GetStringCharsAndLength(cx, str, &length))) {
             return false;
         }
 
-        length = JS_GetStringLength(str);
+        if (!length) {
+            if (useAllocator) {
+                *((const nsACString**)d) = &EmptyCString();
+            } else {
+                nsCString* rs = *((nsCString**)d);
+                rs->Truncate();
+            }
+            return true;
+        }
 
         nsCString *rs;
         if (useAllocator) {
@@ -661,15 +652,13 @@ XPCConvert::JSData2Native(void* d, HandleValue s,
         } else {
             rs = *((nsCString**)d);
         }
-        const char16_t* start = (const char16_t*)chars;
-        const char16_t* end = start + length;
-        CopyUTF16toUTF8(nsDependentSubstring(start, end), *rs);
+        CopyUTF16toUTF8(Substring(chars, length), *rs);
         return true;
     }
 
     case nsXPTType::T_CSTRING:
     {
-        if (JSVAL_IS_NULL(s) || JSVAL_IS_VOID(s)) {
+        if (s.isNull() || s.isUndefined()) {
             if (useAllocator) {
                 nsACString *rs = new nsCString();
                 if (!rs)
@@ -696,6 +685,16 @@ XPCConvert::JSData2Native(void* d, HandleValue s,
             return false;
         }
 
+        if (!length) {
+            if (useAllocator) {
+                *((const nsACString**)d) = &EmptyCString();
+            } else {
+                nsCString* rs = *((nsCString**)d);
+                rs->Truncate();
+            }
+            return true;
+        }
+
         nsACString *rs;
         if (useAllocator) {
             rs = new nsCString();
@@ -721,15 +720,15 @@ XPCConvert::JSData2Native(void* d, HandleValue s,
         MOZ_ASSERT(iid,"can't do interface conversions without iid");
 
         if (iid->Equals(NS_GET_IID(nsIVariant))) {
-            XPCVariant* variant = XPCVariant::newVariant(cx, s);
+            nsCOMPtr<nsIVariant> variant = XPCVariant::newVariant(cx, s);
             if (!variant)
                 return false;
-            *((nsISupports**)d) = static_cast<nsIVariant*>(variant);
+
+            variant.forget(static_cast<nsISupports**>(d));
             return true;
-        } else if (iid->Equals(NS_GET_IID(nsIAtom)) &&
-                   JSVAL_IS_STRING(s)) {
+        } else if (iid->Equals(NS_GET_IID(nsIAtom)) && s.isString()) {
             // We're trying to pass a string as an nsIAtom.  Let's atomize!
-            JSString* str = JSVAL_TO_STRING(s);
+            JSString* str = s.toString();
             const char16_t* chars = JS_GetStringCharsZ(cx, str);
             if (!chars) {
                 if (pErr)
@@ -771,11 +770,11 @@ CreateHolderIfNeeded(HandleObject obj, MutableHandleValue d,
                      nsIXPConnectJSObjectHolder** dest)
 {
     if (dest) {
-        XPCJSObjectHolder* objHolder = XPCJSObjectHolder::newHolder(obj);
+        nsRefPtr<XPCJSObjectHolder> objHolder = XPCJSObjectHolder::newHolder(obj);
         if (!objHolder)
             return false;
 
-        NS_ADDREF(*dest = objHolder);
+        objHolder.forget(dest);
     }
 
     d.setObjectOrNull(obj);
@@ -826,24 +825,18 @@ XPCConvert::NativeInterface2JSObject(MutableHandleValue d,
     // object will create (and fill the cache) from its WrapObject call.
     nsWrapperCache *cache = aHelper.GetWrapperCache();
 
-    RootedObject flat(cx);
-    if (cache) {
-        flat = cache->GetWrapper();
-        if (cache->IsDOMBinding()) {
-            if (!flat) {
-                RootedObject global(cx, xpcscope->GetGlobalJSObject());
-                flat = cache->WrapObject(cx, global);
-                if (!flat)
-                    return false;
-            }
-
-            if (allowNativeWrapper && !JS_WrapObject(cx, &flat))
-                return false;
-
-            return CreateHolderIfNeeded(flat, d, dest);
-        }
-    } else {
-        flat = nullptr;
+    RootedObject flat(cx, cache ? cache->GetWrapper() : nullptr);
+    if (!flat && cache && cache->IsDOMBinding()) {
+        RootedObject global(cx, xpcscope->GetGlobalJSObject());
+        js::AssertSameCompartment(cx, global);
+        flat = cache->WrapObject(cx);
+        if (!flat)
+            return false;
+    }
+    if (flat) {
+        if (allowNativeWrapper && !JS_WrapObject(cx, &flat))
+            return false;
+        return CreateHolderIfNeeded(flat, d, dest);
     }
 
     // Don't double wrap CPOWs. This is a temporary measure for compatibility
@@ -918,7 +911,7 @@ XPCConvert::NativeInterface2JSObject(MutableHandleValue d,
     if (!allowNativeWrapper) {
         d.set(v);
         if (dest)
-            *dest = strongWrapper.forget().get();
+            strongWrapper.forget(dest);
         if (pErr)
             *pErr = NS_OK;
         return true;
@@ -935,14 +928,14 @@ XPCConvert::NativeInterface2JSObject(MutableHandleValue d,
     if (dest) {
         // The strongWrapper still holds the original flat object.
         if (flat == original) {
-            *dest = strongWrapper.forget().get();
+            strongWrapper.forget(dest);
         } else {
             nsRefPtr<XPCJSObjectHolder> objHolder =
                 XPCJSObjectHolder::newHolder(flat);
             if (!objHolder)
                 return false;
 
-            *dest = objHolder.forget().get();
+            objHolder.forget(dest);
         }
     }
 
@@ -1129,9 +1122,9 @@ XPCConvert::JSValToXPCException(MutableHandleValue s,
     AutoJSContext cx;
     AutoExceptionRestorer aer(cx, s);
 
-    if (!JSVAL_IS_PRIMITIVE(s)) {
+    if (!s.isPrimitive()) {
         // we have a JSObject
-        RootedObject obj(cx, JSVAL_TO_OBJECT(s));
+        RootedObject obj(cx, s.toObjectOrNull());
 
         if (!obj) {
             NS_ERROR("when is an object not an object?");
@@ -1149,9 +1142,8 @@ XPCConvert::JSValToXPCException(MutableHandleValue s,
             nsCOMPtr<nsIException> iface = do_QueryInterface(supports);
             if (iface) {
                 // just pass through the exception (with extra ref and all)
-                nsIException* temp = iface;
-                NS_ADDREF(temp);
-                *exceptn = temp;
+                nsCOMPtr<nsIException> temp = iface;
+                temp.forget(exceptn);
                 return NS_OK;
             } else {
                 // it is a wrapped native, but not an exception!
@@ -1218,26 +1210,26 @@ XPCConvert::JSValToXPCException(MutableHandleValue s,
         }
     }
 
-    if (JSVAL_IS_VOID(s) || JSVAL_IS_NULL(s)) {
+    if (s.isUndefined() || s.isNull()) {
         return ConstructException(NS_ERROR_XPC_JS_THREW_NULL,
                                   nullptr, ifaceName, methodName, nullptr,
                                   exceptn, cx, s.address());
     }
 
-    if (JSVAL_IS_NUMBER(s)) {
+    if (s.isNumber()) {
         // lets see if it looks like an nsresult
         nsresult rv;
         double number;
         bool isResult = false;
 
-        if (JSVAL_IS_INT(s)) {
-            rv = (nsresult) JSVAL_TO_INT(s);
+        if (s.isInt32()) {
+            rv = (nsresult) s.toInt32();
             if (NS_FAILED(rv))
                 isResult = true;
             else
-                number = (double) JSVAL_TO_INT(s);
+                number = (double) s.toInt32();
         } else {
-            number = JSVAL_TO_DOUBLE(s);
+            number = s.toDouble();
             if (number > 0.0 &&
                 number < (double)0xffffffff &&
                 0.0 == fmod(number,1)) {
@@ -1362,7 +1354,7 @@ XPCConvert::NativeArray2JS(MutableHandleValue d, const void** s,
 
     // XXX add support to indicate *which* array element was not convertable
 
-    RootedObject array(cx, JS_NewArrayObject(cx, count, nullptr));
+    RootedObject array(cx, JS_NewArrayObject(cx, count));
     if (!array)
         return false;
 
@@ -1766,7 +1758,7 @@ XPCConvert::JSStringWithSize2Native(void* d, HandleValue s,
                                     uint32_t count, const nsXPTType& type,
                                     nsresult* pErr)
 {
-    NS_PRECONDITION(!JSVAL_IS_NULL(s), "bad param");
+    NS_PRECONDITION(!s.isNull(), "bad param");
     NS_PRECONDITION(d, "bad param");
 
     AutoJSContext cx;
@@ -1778,7 +1770,7 @@ XPCConvert::JSStringWithSize2Native(void* d, HandleValue s,
     switch (type.TagPart()) {
         case nsXPTType::T_PSTRING_SIZE_IS:
         {
-            if (JSVAL_IS_VOID(s) || JSVAL_IS_NULL(s)) {
+            if (s.isUndefined() || s.isNull()) {
                 if (0 != count) {
                     if (pErr)
                         *pErr = NS_ERROR_XPC_NOT_ENOUGH_CHARS_IN_STRING;
@@ -1832,7 +1824,7 @@ XPCConvert::JSStringWithSize2Native(void* d, HandleValue s,
             const jschar* chars=nullptr;
             JSString* str;
 
-            if (JSVAL_IS_VOID(s) || JSVAL_IS_NULL(s)) {
+            if (s.isUndefined() || s.isNull()) {
                 if (0 != count) {
                     if (pErr)
                         *pErr = NS_ERROR_XPC_NOT_ENOUGH_CHARS_IN_STRING;
