@@ -6,7 +6,14 @@
 
 /* rendering objects for replaced elements implemented by a plugin */
 
+#include "nsObjectFrame.h"
+
+#include "gfx2DGlue.h"
+#include "mozilla/BasicEvents.h"
+#ifdef XP_WIN
+// This is needed for DoublePassRenderingEvent.
 #include "mozilla/plugins/PluginMessageUtils.h"
+#endif
 
 #include "nscore.h"
 #include "nsCOMPtr.h"
@@ -15,90 +22,38 @@
 #include "nsWidgetsCID.h"
 #include "nsView.h"
 #include "nsViewManager.h"
-#include "nsIDOMEventListener.h"
-#include "nsIDOMDragEvent.h"
-#include "nsPluginHost.h"
 #include "nsString.h"
-#include "nsReadableUtils.h"
 #include "nsGkAtoms.h"
-#include "nsIAppShell.h"
-#include "nsIDocument.h"
-#include "nsINodeInfo.h"
-#include "nsIURL.h"
-#include "nsNetUtil.h"
 #include "nsIPluginInstanceOwner.h"
 #include "nsNPAPIPluginInstance.h"
-#include "nsIPluginTagInfo.h"
-#include "plstr.h"
-#include "nsILinkHandler.h"
-#include "nsIScrollPositionListener.h"
-#include "nsITimer.h"
-#include "nsIDocShellTreeItem.h"
-#include "nsIDocShellTreeOwner.h"
-#include "nsDocShellCID.h"
-#include "nsIWebBrowserChrome.h"
 #include "nsIDOMElement.h"
-#include "nsIDOMNodeList.h"
-#include "nsIDOMHTMLObjectElement.h"
-#include "nsIDOMHTMLEmbedElement.h"
-#include "nsIDOMHTMLAppletElement.h"
-#include "nsIDOMWindow.h"
-#include "nsIDOMEventTarget.h"
-#include "nsIDocumentEncoder.h"
-#include "nsXPIDLString.h"
-#include "nsIDOMRange.h"
-#include "nsIPluginWidget.h"
-#include "nsGUIEvent.h"
 #include "nsRenderingContext.h"
 #include "npapi.h"
-#include "nsTransform2D.h"
-#include "nsIImageLoadingContent.h"
 #include "nsIObjectLoadingContent.h"
-#include "nsPIDOMWindow.h"
 #include "nsContentUtils.h"
 #include "nsDisplayList.h"
-#include "nsAttrName.h"
-#include "nsDataHashtable.h"
-#include "nsDOMClassInfo.h"
 #include "nsFocusManager.h"
 #include "nsLayoutUtils.h"
 #include "nsFrameManager.h"
-#include "nsComponentManagerUtils.h"
 #include "nsIObserverService.h"
-#include "nsIScrollableFrame.h"
-#include "mozilla/Preferences.h"
 #include "GeckoProfiler.h"
 #include <algorithm>
 
-// headers for plugin scriptability
-#include "nsIScriptGlobalObject.h"
-#include "nsIScriptContext.h"
-#include "nsIXPConnect.h"
-#include "nsIXPCScriptable.h"
-#include "nsIClassInfo.h"
-#include "nsIDOMClientRect.h"
-
-#include "nsObjectFrame.h"
 #include "nsIObjectFrame.h"
 #include "nsPluginNativeWindow.h"
-#include "nsIPluginDocument.h"
 #include "FrameLayerBuilder.h"
 
-#include "nsThreadUtils.h"
-
-#include "gfxContext.h"
-#include "gfxPlatform.h"
 #include "ImageLayers.h"
+#include "nsPluginInstanceOwner.h"
 
 #ifdef XP_WIN
 #include "gfxWindowsNativeDrawing.h"
 #include "gfxWindowsSurface.h"
 #endif
 
-#include "gfxImageSurface.h"
-#include "gfxUtils.h"
 #include "Layers.h"
 #include "ReadbackLayer.h"
+#include "ImageContainer.h"
 
 // accessibility support
 #ifdef ACCESSIBILITY
@@ -110,24 +65,12 @@
 #endif /* MOZ_LOGGING */
 #include "prlog.h"
 
-#include <errno.h>
-
-#include "nsContentCID.h"
-static NS_DEFINE_CID(kAppShellCID, NS_APPSHELL_CID);
-
 #ifdef XP_MACOSX
 #include "gfxQuartzNativeDrawing.h"
 #include "nsPluginUtilsOSX.h"
 #include "mozilla/gfx/QuartzSupport.h"
 #endif
 
-#ifdef MOZ_WIDGET_GTK2
-#include <gdk/gdk.h>
-#include <gdk/gdkx.h>
-#include <gtk/gtk.h>
-#include "gfxXlibNativeRenderer.h"
-#endif
-
 #ifdef MOZ_X11
 #include "mozilla/X11Util.h"
 using mozilla::DefaultXDisplay;
@@ -138,13 +81,6 @@ using mozilla::DefaultXDisplay;
 #include <winuser.h>
 #endif
 
-#ifdef XP_OS2
-#define INCL_PM
-#define INCL_GPI
-#include <os2.h>
-#include "gfxOS2Surface.h"
-#endif
-
 #ifdef MOZ_WIDGET_ANDROID
 #include "AndroidBridge.h"
 #include "GLContext.h"
@@ -189,8 +125,8 @@ extern "C" {
     GWorldPtr *   offscreenGWorld,
     UInt32        PixelFormat,
     const Rect *  boundsRect,
-    CTabHandle    cTable,                /* can be NULL */
-    GDHandle      aGDevice,              /* can be NULL */
+    CTabHandle    cTable,                /* can be nullptr */
+    GDHandle      aGDevice,              /* can be nullptr */
     GWorldFlags   flags,
     Ptr           newBuffer,
     SInt32        rowBytes)
@@ -203,7 +139,7 @@ extern "C" {
 #endif /* #if defined(XP_MACOSX) && !defined(__LP64__) */
 
 using namespace mozilla;
-using namespace mozilla::plugins;
+using namespace mozilla::gfx;
 using namespace mozilla::layers;
 
 class PluginBackgroundSink : public ReadbackSink {
@@ -289,9 +225,9 @@ NS_IMETHODIMP nsObjectFrame::GetPluginPort(HWND *aPort)
 #endif
 
 void
-nsObjectFrame::Init(nsIContent*      aContent,
-                    nsIFrame*        aParent,
-                    nsIFrame*        aPrevInFlow)
+nsObjectFrame::Init(nsIContent*       aContent,
+                    nsContainerFrame* aParent,
+                    nsIFrame*         aPrevInFlow)
 {
   PR_LOG(GetObjectFrameLog(), PR_LOG_DEBUG,
          ("Initializing nsObjectFrame %p for content %p\n", this, aContent));
@@ -310,6 +246,9 @@ nsObjectFrame::DestroyFrom(nsIFrame* aDestructRoot)
   nsCOMPtr<nsIObjectLoadingContent> objContent(do_QueryInterface(mContent));
   NS_ASSERTION(objContent, "Why not an object loading content?");
 
+  // The content might not have a reference to the instance owner any longer in
+  // the case of re-entry during instantiation or teardown, so make sure we're
+  // dissociated.
   if (mInstanceOwner) {
     mInstanceOwner->SetFrame(nullptr);
   }
@@ -319,8 +258,6 @@ nsObjectFrame::DestroyFrom(nsIFrame* aDestructRoot)
     mBackgroundSink->Destroy();
   }
 
-  SetInstanceOwner(nullptr);
-
   nsObjectFrameSuper::DestroyFrom(aDestructRoot);
 }
 
@@ -346,8 +283,8 @@ nsObjectFrame::GetType() const
   return nsGkAtoms::objectFrame; 
 }
 
-#ifdef DEBUG
-NS_IMETHODIMP
+#ifdef DEBUG_FRAME_DUMP
+nsresult
 nsObjectFrame::GetFrameName(nsAString& aResult) const
 {
   return MakeFrameName(NS_LITERAL_STRING("ObjectFrame"), aResult);
@@ -420,8 +357,7 @@ nsObjectFrame::PrepForDrawing(nsIWidget *aWidget)
     configuration->mBounds.height = NSAppUnitsToIntPixels(mRect.height, appUnitsPerDevPixel);
     parentWidget->ConfigureChildren(configurations);
 
-    nsRefPtr<nsDeviceContext> dx;
-    viewMan->GetDeviceContext(*getter_AddRefs(dx));
+    nsRefPtr<nsDeviceContext> dx = viewMan->GetDeviceContext();
     mInnerView->AttachWidgetEventHandler(mWidget);
 
 #ifdef XP_MACOSX
@@ -501,57 +437,57 @@ nsObjectFrame::GetDesiredSize(nsPresContext* aPresContext,
                               nsHTMLReflowMetrics& aMetrics)
 {
   // By default, we have no area
-  aMetrics.width = 0;
-  aMetrics.height = 0;
+  aMetrics.Width() = 0;
+  aMetrics.Height() = 0;
 
   if (IsHidden(false)) {
     return;
   }
   
-  aMetrics.width = aReflowState.ComputedWidth();
-  aMetrics.height = aReflowState.ComputedHeight();
+  aMetrics.Width() = aReflowState.ComputedWidth();
+  aMetrics.Height() = aReflowState.ComputedHeight();
 
   // for EMBED and APPLET, default to 240x200 for compatibility
   nsIAtom *atom = mContent->Tag();
   if (atom == nsGkAtoms::applet || atom == nsGkAtoms::embed) {
-    if (aMetrics.width == NS_UNCONSTRAINEDSIZE) {
-      aMetrics.width = clamped(nsPresContext::CSSPixelsToAppUnits(EMBED_DEF_WIDTH),
-                               aReflowState.mComputedMinWidth,
-                               aReflowState.mComputedMaxWidth);
+    if (aMetrics.Width() == NS_UNCONSTRAINEDSIZE) {
+      aMetrics.Width() = clamped(nsPresContext::CSSPixelsToAppUnits(EMBED_DEF_WIDTH),
+                               aReflowState.ComputedMinWidth(),
+                               aReflowState.ComputedMaxWidth());
     }
-    if (aMetrics.height == NS_UNCONSTRAINEDSIZE) {
-      aMetrics.height = clamped(nsPresContext::CSSPixelsToAppUnits(EMBED_DEF_HEIGHT),
-                                aReflowState.mComputedMinHeight,
-                                aReflowState.mComputedMaxHeight);
+    if (aMetrics.Height() == NS_UNCONSTRAINEDSIZE) {
+      aMetrics.Height() = clamped(nsPresContext::CSSPixelsToAppUnits(EMBED_DEF_HEIGHT),
+                                aReflowState.ComputedMinHeight(),
+                                aReflowState.ComputedMaxHeight());
     }
 
-#if defined (MOZ_WIDGET_GTK2)
+#if defined(MOZ_WIDGET_GTK)
     // We need to make sure that the size of the object frame does not
     // exceed the maximum size of X coordinates.  See bug #225357 for
     // more information.  In theory Gtk2 can handle large coordinates,
     // but underlying plugins can't.
-    aMetrics.height = std::min(aPresContext->DevPixelsToAppUnits(INT16_MAX), aMetrics.height);
-    aMetrics.width = std::min(aPresContext->DevPixelsToAppUnits(INT16_MAX), aMetrics.width);
+    aMetrics.Height() = std::min(aPresContext->DevPixelsToAppUnits(INT16_MAX), aMetrics.Height());
+    aMetrics.Width() = std::min(aPresContext->DevPixelsToAppUnits(INT16_MAX), aMetrics.Width());
 #endif
   }
 
   // At this point, the width has an unconstrained value only if we have
   // nothing to go on (no width set, no information from the plugin, nothing).
   // Make up a number.
-  if (aMetrics.width == NS_UNCONSTRAINEDSIZE) {
-    aMetrics.width =
-      (aReflowState.mComputedMinWidth != NS_UNCONSTRAINEDSIZE) ?
-        aReflowState.mComputedMinWidth : 0;
+  if (aMetrics.Width() == NS_UNCONSTRAINEDSIZE) {
+    aMetrics.Width() =
+      (aReflowState.ComputedMinWidth() != NS_UNCONSTRAINEDSIZE) ?
+        aReflowState.ComputedMinWidth() : 0;
   }
 
   // At this point, the height has an unconstrained value only in two cases:
   // a) We are in standards mode with percent heights and parent is auto-height
   // b) We have no height information at all.
   // In either case, we have to make up a number.
-  if (aMetrics.height == NS_UNCONSTRAINEDSIZE) {
-    aMetrics.height =
-      (aReflowState.mComputedMinHeight != NS_UNCONSTRAINEDSIZE) ?
-        aReflowState.mComputedMinHeight : 0;
+  if (aMetrics.Height() == NS_UNCONSTRAINEDSIZE) {
+    aMetrics.Height() =
+      (aReflowState.ComputedMinHeight() != NS_UNCONSTRAINEDSIZE) ?
+        aReflowState.ComputedMinHeight() : 0;
   }
 
   // XXXbz don't add in the border and padding, because we screw up our
@@ -561,7 +497,7 @@ nsObjectFrame::GetDesiredSize(nsPresContext* aPresContext,
   // call the superclass in all cases.
 }
 
-NS_IMETHODIMP
+void
 nsObjectFrame::Reflow(nsPresContext*           aPresContext,
                       nsHTMLReflowMetrics&     aMetrics,
                       const nsHTMLReflowState& aReflowState,
@@ -580,17 +516,17 @@ nsObjectFrame::Reflow(nsPresContext*           aPresContext,
   // plugin needs to see that haven't arrived yet.
   if (!GetContent()->IsDoneAddingChildren()) {
     aStatus = NS_FRAME_COMPLETE;
-    return NS_OK;
+    return;
   }
 
   // if we are printing or print previewing, bail for now
   if (aPresContext->Medium() == nsGkAtoms::print) {
     aStatus = NS_FRAME_COMPLETE;
-    return NS_OK;
+    return;
   }
 
-  nsRect r(0, 0, aMetrics.width, aMetrics.height);
-  r.Deflate(aReflowState.mComputedBorderPadding);
+  nsRect r(0, 0, aMetrics.Width(), aMetrics.Height());
+  r.Deflate(aReflowState.ComputedPhysicalBorderPadding());
 
   if (mInnerView) {
     nsViewManager* vm = mInnerView->GetViewManager();
@@ -607,7 +543,6 @@ nsObjectFrame::Reflow(nsPresContext*           aPresContext,
   aStatus = NS_FRAME_COMPLETE;
 
   NS_FRAME_SET_TRUNCATION(aStatus, aReflowState, aMetrics);
-  return NS_OK;
 }
 
 ///////////// nsIReflowCallback ///////////////
@@ -780,6 +715,10 @@ nsObjectFrame::UnregisterPluginForGeometryUpdates()
 void
 nsObjectFrame::SetInstanceOwner(nsPluginInstanceOwner* aOwner)
 {
+  // The ownership model here is historically fuzzy. This should only be called
+  // by nsPluginInstanceOwner when it is given a new frame, and
+  // nsObjectLoadingContent should be arbitrating frame-ownership via its
+  // HasNewFrame callback.
   mInstanceOwner = aOwner;
   if (mInstanceOwner) {
     return;
@@ -861,7 +800,7 @@ nsIntPoint nsObjectFrame::GetWindowOriginInPixels(bool aWindowless)
                     PresContext()->AppUnitsToDevPixels(origin.y));
 }
 
-NS_IMETHODIMP
+void
 nsObjectFrame::DidReflow(nsPresContext*            aPresContext,
                          const nsHTMLReflowState*  aReflowState,
                          nsDidReflowStatus         aStatus)
@@ -875,12 +814,12 @@ nsObjectFrame::DidReflow(nsPresContext*            aPresContext,
     objContent->HasNewFrame(this);
   }
 
-  nsresult rv = nsObjectFrameSuper::DidReflow(aPresContext, aReflowState, aStatus);
+  nsObjectFrameSuper::DidReflow(aPresContext, aReflowState, aStatus);
 
   // The view is created hidden; once we have reflowed it and it has been
   // positioned then we show it.
-  if (aStatus != nsDidReflowStatus::FINISHED) 
-    return rv;
+  if (aStatus != nsDidReflowStatus::FINISHED)
+    return;
 
   if (HasView()) {
     nsView* view = GetView();
@@ -888,8 +827,6 @@ nsObjectFrame::DidReflow(nsPresContext*            aPresContext,
     if (vm)
       vm->SetViewVisibility(view, IsHidden() ? nsViewVisibility_kHide : nsViewVisibility_kShow);
   }
-
-  return rv;
 }
 
 /* static */ void
@@ -915,23 +852,24 @@ public:
   }
 #endif
 
-  virtual nsRect GetBounds(nsDisplayListBuilder* aBuilder, bool* aSnap);
+  virtual nsRect GetBounds(nsDisplayListBuilder* aBuilder,
+                           bool* aSnap) MOZ_OVERRIDE;
   virtual bool ComputeVisibility(nsDisplayListBuilder* aBuilder,
                                    nsRegion* aVisibleRegion,
-                                   const nsRect& aAllowVisibleRegionExpansion);
+                                   const nsRect& aAllowVisibleRegionExpansion) MOZ_OVERRIDE;
 
   NS_DISPLAY_DECL_NAME("PluginReadback", TYPE_PLUGIN_READBACK)
 
   virtual already_AddRefed<Layer> BuildLayer(nsDisplayListBuilder* aBuilder,
                                              LayerManager* aManager,
-                                             const ContainerParameters& aContainerParameters)
+                                             const ContainerLayerParameters& aContainerParameters) MOZ_OVERRIDE
   {
     return static_cast<nsObjectFrame*>(mFrame)->BuildLayer(aBuilder, aManager, this, aContainerParameters);
   }
 
   virtual LayerState GetLayerState(nsDisplayListBuilder* aBuilder,
                                    LayerManager* aManager,
-                                   const ContainerParameters& aParameters)
+                                   const ContainerLayerParameters& aParameters) MOZ_OVERRIDE
   {
     return LAYER_ACTIVE;
   }
@@ -985,23 +923,24 @@ public:
   }
 #endif
 
-  virtual nsRect GetBounds(nsDisplayListBuilder* aBuilder, bool* aSnap);
+  virtual nsRect GetBounds(nsDisplayListBuilder* aBuilder,
+                           bool* aSnap) MOZ_OVERRIDE;
   virtual bool ComputeVisibility(nsDisplayListBuilder* aBuilder,
                                    nsRegion* aVisibleRegion,
-                                   const nsRect& aAllowVisibleRegionExpansion);
+                                   const nsRect& aAllowVisibleRegionExpansion) MOZ_OVERRIDE;
 
   NS_DISPLAY_DECL_NAME("PluginVideo", TYPE_PLUGIN_VIDEO)
 
   virtual already_AddRefed<Layer> BuildLayer(nsDisplayListBuilder* aBuilder,
                                              LayerManager* aManager,
-                                             const ContainerParameters& aContainerParameters)
+                                             const ContainerLayerParameters& aContainerParameters) MOZ_OVERRIDE
   {
     return static_cast<nsObjectFrame*>(mFrame)->BuildLayer(aBuilder, aManager, this, aContainerParameters);
   }
 
   virtual LayerState GetLayerState(nsDisplayListBuilder* aBuilder,
                                    LayerManager* aManager,
-                                   const ContainerParameters& aParameters)
+                                   const ContainerLayerParameters& aParameters) MOZ_OVERRIDE
   {
     return LAYER_ACTIVE;
   }
@@ -1062,9 +1001,8 @@ nsDisplayPlugin::ComputeVisibility(nsDisplayListBuilder* aBuilder,
         ReferenceFrame()->PresContext()->AppUnitsPerDevPixel();
       f->mNextConfigurationBounds = rAncestor.ToNearestPixels(appUnitsPerDevPixel);
 
-      bool snap;
       nsRegion visibleRegion;
-      visibleRegion.And(*aVisibleRegion, GetBounds(aBuilder, &snap));
+      visibleRegion.And(*aVisibleRegion, GetClippedBounds(aBuilder));
       // Make visibleRegion relative to f
       visibleRegion.MoveBy(-ToReferenceFrame());
 
@@ -1233,8 +1171,6 @@ nsObjectFrame::BuildDisplayList(nsDisplayListBuilder*   aBuilder,
   }
 #endif
 
-  nsDisplayList replacedContent;
-
   if (aBuilder->IsForPainting() && mInstanceOwner && mInstanceOwner->UseAsyncRendering()) {
     NPWindow* window = nullptr;
     mInstanceOwner->GetWindow(window);
@@ -1248,9 +1184,12 @@ nsObjectFrame::BuildDisplayList(nsDisplayListBuilder*   aBuilder,
     mInstanceOwner->NotifyPaintWaiter(aBuilder);
   }
 
+  DisplayListClipState::AutoClipContainingBlockDescendantsToContentBox
+    clip(aBuilder, this);
+
   // determine if we are printing
   if (type == nsPresContext::eContext_Print) {
-    replacedContent.AppendNewToTop(new (aBuilder)
+    aLists.Content()->AppendNewToTop(new (aBuilder)
       nsDisplayGeneric(aBuilder, this, PaintPrintPlugin, "PrintPlugin",
                        nsDisplayItem::TYPE_PRINT_PLUGIN));
   } else {
@@ -1264,7 +1203,7 @@ nsObjectFrame::BuildDisplayList(nsDisplayListBuilder*   aBuilder,
     if (aBuilder->IsPaintingToWindow() &&
         state == LAYER_ACTIVE &&
         IsTransparentMode()) {
-      replacedContent.AppendNewToTop(new (aBuilder)
+      aLists.Content()->AppendNewToTop(new (aBuilder)
         nsDisplayPluginReadback(aBuilder, this));
     }
 #endif
@@ -1277,32 +1216,17 @@ nsObjectFrame::BuildDisplayList(nsDisplayListBuilder*   aBuilder,
       mInstanceOwner->GetVideos(videos);
 
       for (uint32_t i = 0; i < videos.Length(); i++) {
-        replacedContent.AppendNewToTop(new (aBuilder)
+        aLists.Content()->AppendNewToTop(new (aBuilder)
           nsDisplayPluginVideo(aBuilder, this, videos[i]));
       }
     }
 #endif
 
-    replacedContent.AppendNewToTop(new (aBuilder)
+    aLists.Content()->AppendNewToTop(new (aBuilder)
       nsDisplayPlugin(aBuilder, this));
   }
-
-  WrapReplacedContentForBorderRadius(aBuilder, &replacedContent, aLists);
 }
 
-#ifdef XP_OS2
-static void *
-GetPSFromRC(nsRenderingContext& aRenderingContext)
-{
-  nsRefPtr<gfxASurface>
-    surf = aRenderingContext.ThebesContext()->CurrentSurface();
-  if (!surf || surf->CairoStatus())
-    return nullptr;
-  return (void *)(static_cast<gfxOS2Surface*>
-                  (static_cast<gfxASurface*>(surf.get()))->GetPS());
-}
-#endif
-
 void
 nsObjectFrame::PrintPlugin(nsRenderingContext& aRenderingContext,
                            const nsRect& aDirtyRect)
@@ -1404,7 +1328,8 @@ nsObjectFrame::PrintPlugin(nsRenderingContext& aRenderingContext,
     return;
   }
   GWorldPtr gWorld;
-  if (::NewGWorldFromPtr(&gWorld, k32ARGBPixelFormat, &gwBounds, NULL, NULL, 0,
+  if (::NewGWorldFromPtr(&gWorld, k32ARGBPixelFormat, &gwBounds,
+                         nullptr, nullptr, 0,
                          buffer.Elements(), window.width * 4) != noErr) {
     ::CGContextRelease(cgBuffer);
     nativeDraw.EndNativeDrawing();
@@ -1455,15 +1380,6 @@ nsObjectFrame::PrintPlugin(nsRenderingContext& aRenderingContext,
   (void)window;
   (void)npprint;
 
-#elif defined(XP_OS2)
-  void *hps = GetPSFromRC(aRenderingContext);
-  if (!hps)
-    return;
-
-  npprint.print.embedPrint.platformPrint = hps;
-  npprint.print.embedPrint.window = window;
-  // send off print info to plugin
-  pi->Print(&npprint);
 #elif defined(XP_WIN)
 
   /* On Windows, we use the win32 printing surface to print.  This, in
@@ -1540,20 +1456,6 @@ nsObjectFrame::GetPaintedRect(nsDisplayPlugin* aItem)
   return r;
 }
 
-void
-nsObjectFrame::UpdateImageLayer(const gfxRect& aRect)
-{
-  if (!mInstanceOwner) {
-    return;
-  }
-
-#ifdef XP_MACOSX
-  if (!mInstanceOwner->UseAsyncRendering()) {
-    mInstanceOwner->DoCocoaEventDrawRect(aRect, nullptr);
-  }
-#endif
-}
-
 LayerState
 nsObjectFrame::GetLayerState(nsDisplayListBuilder* aBuilder,
                              LayerManager* aManager)
@@ -1561,14 +1463,6 @@ nsObjectFrame::GetLayerState(nsDisplayListBuilder* aBuilder,
   if (!mInstanceOwner)
     return LAYER_NONE;
 
-#ifdef XP_MACOSX
-  if (!mInstanceOwner->UseAsyncRendering() &&
-      mInstanceOwner->IsRemoteDrawingCoreAnimation() &&
-      mInstanceOwner->GetEventModel() == NPEventModelCocoa) {
-    return LAYER_ACTIVE;
-  }
-#endif
-
 #ifdef MOZ_WIDGET_ANDROID
   // We always want a layer on Honeycomb and later
   if (AndroidBridge::Bridge()->GetAPIVersion() >= 11)
@@ -1586,7 +1480,7 @@ already_AddRefed<Layer>
 nsObjectFrame::BuildLayer(nsDisplayListBuilder* aBuilder,
                           LayerManager* aManager,
                           nsDisplayItem* aItem,
-                          const ContainerParameters& aContainerParameters)
+                          const ContainerLayerParameters& aContainerParameters)
 {
   if (!mInstanceOwner)
     return nullptr;
@@ -1605,7 +1499,7 @@ nsObjectFrame::BuildLayer(nsDisplayListBuilder* aBuilder,
     scaleFactor = 1.0;
   }
   int intScaleFactor = ceil(scaleFactor);
-  gfxIntSize size(window->width * intScaleFactor, window->height * intScaleFactor);
+  IntSize size(window->width * intScaleFactor, window->height * intScaleFactor);
 
   nsRect area = GetContentRectRelativeToSelf() + aItem->ToReferenceFrame();
   gfxRect r = nsLayoutUtils::RectToGfxRect(area, PresContext()->AppUnitsPerDevPixel());
@@ -1632,16 +1526,20 @@ nsObjectFrame::BuildLayer(nsDisplayListBuilder* aBuilder,
 
     NS_ASSERTION(layer->GetType() == Layer::TYPE_IMAGE, "Bad layer type");
     ImageLayer* imglayer = static_cast<ImageLayer*>(layer.get());
-    UpdateImageLayer(r);
+#ifdef XP_MACOSX
+    if (!mInstanceOwner->UseAsyncRendering()) {
+      mInstanceOwner->DoCocoaEventDrawRect(r, nullptr);
+    }
+#endif
 
-    imglayer->SetScaleToSize(size, ImageLayer::SCALE_STRETCH);
+    imglayer->SetScaleToSize(size, ScaleMode::STRETCH);
     imglayer->SetContainer(container);
-    gfxPattern::GraphicsFilter filter =
+    GraphicsFilter filter =
       nsLayoutUtils::GetGraphicsFilterForFrame(this);
 #ifdef MOZ_GFX_OPTIMIZE_MOBILE
     if (!aManager->IsCompositingCheap()) {
       // Pixman just horrible with bilinear filter scaling
-      filter = gfxPattern::FILTER_NEAREST;
+      filter = GraphicsFilter::FILTER_NEAREST;
     }
 #endif
     imglayer->SetFilter(filter);
@@ -1686,11 +1584,11 @@ nsObjectFrame::BuildLayer(nsDisplayListBuilder* aBuilder,
     NS_ASSERTION(layer->GetType() == Layer::TYPE_READBACK, "Bad layer type");
 
     ReadbackLayer* readback = static_cast<ReadbackLayer*>(layer.get());
-    if (readback->GetSize() != nsIntSize(size.width, size.height)) {
+    if (readback->GetSize() != ThebesIntSize(size)) {
       // This will destroy any old background sink and notify us that the
       // background is now unknown
       readback->SetSink(nullptr);
-      readback->SetSize(nsIntSize(size.width, size.height));
+      readback->SetSize(ThebesIntSize(size));
 
       if (mBackgroundSink) {
         // Maybe we still have a background sink associated with another
@@ -1709,11 +1607,12 @@ nsObjectFrame::BuildLayer(nsDisplayListBuilder* aBuilder,
   }
 
   // Set a transform on the layer to draw the plugin in the right place
-  gfxMatrix transform;
-  transform.Translate(r.TopLeft() + aContainerParameters.mOffset);
+  Matrix transform;
+  gfxPoint p = r.TopLeft() + aContainerParameters.mOffset;
+  transform.Translate(p.x, p.y);
 
-  layer->SetBaseTransform(gfx3DMatrix::From2D(transform));
-  layer->SetVisibleRegion(nsIntRect(0, 0, size.width, size.height));
+  layer->SetBaseTransform(Matrix4x4::From2D(transform));
+  layer->SetVisibleRegion(ThebesIntRect(IntRect(IntPoint(0, 0), size)));
   return layer.forget();
 }
 
@@ -1816,7 +1715,7 @@ nsObjectFrame::PaintPlugin(nsDisplayListBuilder* aBuilder,
 
       // this rect is used only in the CoreGraphics drawing model
       gfxRect tmpRect(0, 0, 0, 0);
-      mInstanceOwner->Paint(tmpRect, NULL);
+      mInstanceOwner->Paint(tmpRect, nullptr);
     }
   }
 #elif defined(MOZ_X11)
@@ -1864,7 +1763,7 @@ nsObjectFrame::PaintPlugin(nsDisplayListBuilder* aBuilder,
         // double pass render. If this plugin isn't oop, the register window message
         // will be ignored.
         NPEvent pluginEvent;
-        pluginEvent.event = DoublePassRenderingEvent();
+        pluginEvent.event = plugins::DoublePassRenderingEvent();
         pluginEvent.wParam = 0;
         pluginEvent.lParam = 0;
         if (pluginEvent.event)
@@ -1937,105 +1836,13 @@ nsObjectFrame::PaintPlugin(nsDisplayListBuilder* aBuilder,
 
     ctx->SetMatrix(currentMatrix);
   }
-#elif defined(XP_OS2)
-  nsRefPtr<nsNPAPIPluginInstance> inst;
-  GetPluginInstance(getter_AddRefs(inst));
-  if (inst) {
-    // Look if it's windowless
-    NPWindow *window;
-    mInstanceOwner->GetWindow(window);
-
-    if (window->type == NPWindowTypeDrawable) {
-      // FIXME - Bug 385435: Doesn't aDirtyRect need translating too?
-      nsRenderingContext::AutoPushTranslation
-        translate(&aRenderingContext, aPluginRect.TopLeft());
-
-      // check if we need to call SetWindow with updated parameters
-      bool doupdatewindow = false;
-      // the offset of the DC
-      nsIntPoint origin;
-
-      /*
-       * Layout now has an optimized way of painting. Now we always get
-       * a new drawing surface, sized to be just what's needed. Windowless
-       * plugins need a transform applied to their origin so they paint
-       * in the right place. Since |SetWindow| is no longer being used
-       * to tell the plugin where it is, we dispatch a NPWindow through
-       * |HandleEvent| to tell the plugin when its window moved
-       */
-      gfxContext *ctx = aRenderingContext.ThebesContext();
-
-      gfxMatrix ctxMatrix = ctx->CurrentMatrix();
-      if (ctxMatrix.HasNonTranslation()) {
-        // soo; in the future, we should be able to render
-        // the object content to an offscreen DC, and then
-        // composite it in with the right transforms.
-
-        // But, we don't bother doing that, because we don't
-        // have the event handling story figured out yet.
-        // Instead, let's just bail.
-
-        return;
-      }
-
-      origin.x = NSToIntRound(ctxMatrix.GetTranslation().x);
-      origin.y = NSToIntRound(ctxMatrix.GetTranslation().y);
-
-      /* Need to force the clip to be set */
-      ctx->UpdateSurfaceClip();
-
-      /* Set the device offsets as appropriate, for whatever our current group offsets might be */
-      gfxFloat xoff, yoff;
-      nsRefPtr<gfxASurface> surf = ctx->CurrentSurface(&xoff, &yoff);
-
-      if (surf->CairoStatus() != 0) {
-        NS_WARNING("Plugin is being asked to render to a surface that's in error!");
-        return;
-      }
-
-      // check if we need to update the PS
-      HPS hps = (HPS)GetPSFromRC(aRenderingContext);
-      if (reinterpret_cast<HPS>(window->window) != hps) {
-        window->window = reinterpret_cast<void*>(hps);
-        doupdatewindow = true;
-      }
-      LONG lPSid = GpiSavePS(hps);
-      RECTL rclViewport;
-      if (GpiQueryDevice(hps) != NULLHANDLE) { // ensure that we have an associated HDC
-        if (GpiQueryPageViewport(hps, &rclViewport)) {
-          rclViewport.xLeft += (LONG)xoff;
-          rclViewport.xRight += (LONG)xoff;
-          rclViewport.yBottom += (LONG)yoff;
-          rclViewport.yTop += (LONG)yoff;
-          GpiSetPageViewport(hps, &rclViewport);
-        }
-      }
-
-      if ((window->x != origin.x) || (window->y != origin.y)) {
-        window->x = origin.x;
-        window->y = origin.y;
-        doupdatewindow = true;
-      }
-
-      // if our location or visible area has changed, we need to tell the plugin
-      if (doupdatewindow) {
-        inst->SetWindow(window);        
-      }
-
-      mInstanceOwner->Paint(aDirtyRect, hps);
-      if (lPSid >= 1) {
-        GpiRestorePS(hps, lPSid);
-      }
-      surf->MarkDirty();
-    }
-  }
 #endif
 }
 
-NS_IMETHODIMP
+nsresult
 nsObjectFrame::HandleEvent(nsPresContext* aPresContext,
-                           nsGUIEvent*     anEvent,
-                           nsEventStatus*  anEventStatus)
+                           WidgetGUIEvent* anEvent,
+                           nsEventStatus* anEventStatus)
 {
   NS_ENSURE_ARG_POINTER(anEvent);
   NS_ENSURE_ARG_POINTER(anEventStatus);
@@ -2068,7 +1875,7 @@ nsObjectFrame::HandleEvent(nsPresContext* aPresContext,
 #endif
 
   if (mInstanceOwner->SendNativeEvents() &&
-      NS_IS_PLUGIN_EVENT(anEvent)) {
+      anEvent->IsNativeEventDelivererForPlugin()) {
     *anEventStatus = mInstanceOwner->ProcessEvent(*anEvent);
     // Due to plugin code reentering Gecko, this frame may be dead at this
     // point.
@@ -2090,24 +1897,30 @@ nsObjectFrame::HandleEvent(nsPresContext* aPresContext,
     // point.
     return rv;
   }
+
+  // These two calls to nsIPresShell::SetCapturingContext() (on mouse-down
+  // and mouse-up) are needed to make the routing of mouse events while
+  // dragging conform to standard OS X practice, and to the Cocoa NPAPI spec.
+  // See bug 525078 and bug 909678.
+  if (anEvent->message == NS_MOUSE_BUTTON_DOWN) {
+    nsIPresShell::SetCapturingContent(GetContent(), CAPTURE_IGNOREALLOWED);
+  }
 #endif
 
-  return nsObjectFrameSuper::HandleEvent(aPresContext, anEvent, anEventStatus);
-}
+  rv = nsObjectFrameSuper::HandleEvent(aPresContext, anEvent, anEventStatus);
+
+  // We need to be careful from this point because the call to
+  // nsObjectFrameSuper::HandleEvent() might have killed us.
 
 #ifdef XP_MACOSX
-// Needed to make the routing of mouse events while dragging conform to
-// standard OS X practice, and to the Cocoa NPAPI spec.  See bug 525078.
-NS_IMETHODIMP
-nsObjectFrame::HandlePress(nsPresContext* aPresContext,
-                           nsGUIEvent*    anEvent,
-                           nsEventStatus* anEventStatus)
-{
-  nsIPresShell::SetCapturingContent(GetContent(), CAPTURE_IGNOREALLOWED);
-  return nsObjectFrameSuper::HandlePress(aPresContext, anEvent, anEventStatus);
-}
+  if (anEvent->message == NS_MOUSE_BUTTON_UP) {
+    nsIPresShell::SetCapturingContent(nullptr, 0);
+  }
 #endif
 
+  return rv;
+}
+
 nsresult
 nsObjectFrame::GetPluginInstance(nsNPAPIPluginInstance** aPluginInstance)
 {
@@ -2120,7 +1933,7 @@ nsObjectFrame::GetPluginInstance(nsNPAPIPluginInstance** aPluginInstance)
   return mInstanceOwner->GetInstance(aPluginInstance);
 }
 
-NS_IMETHODIMP
+nsresult
 nsObjectFrame::GetCursor(const nsPoint& aPoint, nsIFrame::Cursor& aCursor)
 {
   if (!mInstanceOwner) {
