@@ -2,6 +2,9 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this file,
 # You can obtain one at http://mozilla.org/MPL/2.0/.
 
+import datetime
+import mozlog
+import moznetwork
 import select
 import socket
 import time
@@ -10,7 +13,7 @@ import re
 import posixpath
 import subprocess
 import StringIO
-from devicemanager import DeviceManager, DMError, NetworkTools, _pop_last_line
+from devicemanager import DeviceManager, DMError, _pop_last_line
 import errno
 from distutils.version import StrictVersion
 
@@ -22,7 +25,6 @@ class DeviceManagerSUT(DeviceManager):
     app must be present and listening for connections for this to work.
     """
 
-    debug = 2
     _base_prompt = '$>'
     _base_prompt_re = '\$\>'
     _prompt_sep = '\x00'
@@ -33,7 +35,9 @@ class DeviceManagerSUT(DeviceManager):
     reboot_timeout = 600
     reboot_settling_time = 60
 
-    def __init__(self, host, port = 20701, retryLimit = 5, deviceRoot = None, **kwargs):
+    def __init__(self, host, port = 20701, retryLimit = 5,
+            deviceRoot = None, logLevel = mozlog.ERROR, **kwargs):
+        DeviceManager.__init__(self, logLevel)
         self.host = host
         self.port = port
         self.retryLimit = retryLimit
@@ -130,13 +134,12 @@ class DeviceManagerSUT(DeviceManager):
                 # couldn't execute it). retry otherwise
                 if err.fatal:
                     raise err
-                if self.debug >= 4:
-                    print err
+                self._logger.debug(err)
                 retries += 1
                 # if we lost the connection or failed to establish one, wait a bit
                 if retries < retryLimit and not self._sock:
                     sleep_time = 5 * retries
-                    print 'Could not connect; sleeping for %d seconds.' % sleep_time
+                    self._logger.info('Could not connect; sleeping for %d seconds.' % sleep_time)
                     time.sleep(sleep_time)
 
         raise DMError("Remote Device Error: unable to connect to %s after %s attempts" % (self.host, retryLimit))
@@ -162,8 +165,8 @@ class DeviceManagerSUT(DeviceManager):
 
         if not self._sock:
             try:
-                if self.debug >= 1 and self._everConnected:
-                    print "reconnecting socket"
+                if self._everConnected:
+                    self._logger.info("reconnecting socket")
                 self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
             except socket.error, msg:
                 self._sock = None
@@ -202,13 +205,12 @@ class DeviceManagerSUT(DeviceManager):
                         raise DMError("Remote Device Error: we had %s bytes of data to send, but "
                                       "only sent %s" % (len(cmd['data']), sent))
 
-                if self.debug >= 4:
-                    print "sent cmd: " + str(cmd['cmd'])
+                self._logger.debug("sent cmd: %s" % cmd['cmd'])
             except socket.error, msg:
                 self._sock.close()
                 self._sock = None
-                if self.debug >= 1:
-                    print "Remote Device Error: Error sending data to socket. cmd="+str(cmd['cmd'])+"; err="+str(msg)
+                self._logger.error("Remote Device Error: Error sending data"\
+                        " to socket. cmd=%s; err=%s" % (cmd['cmd'], msg))
                 return False
 
             # Check if the command should close the socket
@@ -226,22 +228,22 @@ class DeviceManagerSUT(DeviceManager):
                     socketClosed = False
                     errStr = ''
                     temp = ''
-                    if self.debug >= 4:
-                        print "recv'ing..."
+                    self._logger.debug("recv'ing...")
 
                     # Get our response
                     try:
-                          # Wait up to a second for socket to become ready for reading...
+                        # Wait up to a second for socket to become ready for reading...
                         if select.select([self._sock], [], [], select_timeout)[0]:
                             temp = self._sock.recv(1024)
-                            if self.debug >= 4:
-                                print "response: " + str(temp)
+                            self._logger.debug("response: %s" % temp)
                             timer = 0
                             if not temp:
                                 socketClosed = True
                                 errStr = 'connection closed'
                         timer += select_timeout
                         if timer > timeout:
+                            self._sock.close()
+                            self._sock = None
                             raise DMError("Automation Error: Timeout in command %s" % cmd['cmd'], fatal=True)
                     except socket.error, err:
                         socketClosed = True
@@ -340,9 +342,10 @@ class DeviceManagerSUT(DeviceManager):
         # woops, we couldn't find an end of line/return value
         raise DMError("Automation Error: Error finding end of line/return value when running '%s'" % cmdline)
 
-    def pushFile(self, localname, destname, retryLimit = None):
+    def pushFile(self, localname, destname, retryLimit=None, createDir=True):
         retryLimit = retryLimit or self.retryLimit
-        self.mkDirs(destname)
+        if createDir:
+            self.mkDirs(destname)
 
         try:
             filesize = os.path.getsize(localname)
@@ -352,8 +355,7 @@ class DeviceManagerSUT(DeviceManager):
         except OSError:
             raise DMError("DeviceManager: Error reading file to push")
 
-        if (self.debug >= 3):
-            print "push returned: %s" % remoteHash
+        self._logger.debug("push returned: %s" % remoteHash)
 
         localHash = self._getLocalHash(localname)
 
@@ -365,22 +367,19 @@ class DeviceManagerSUT(DeviceManager):
         if not self.dirExists(name):
             self._runCmds([{ 'cmd': 'mkdr ' + name }])
 
-    def pushDir(self, localDir, remoteDir, retryLimit = None):
+    def pushDir(self, localDir, remoteDir, retryLimit=None, timeout=None):
         retryLimit = retryLimit or self.retryLimit
-        if (self.debug >= 2):
-            print "pushing directory: %s to %s" % (localDir, remoteDir)
+        self._logger.info("pushing directory: %s to %s" % (localDir, remoteDir))
 
         existentDirectories = []
         for root, dirs, files in os.walk(localDir, followlinks=True):
-            parts = root.split(localDir)
+            _, subpath = root.split(localDir)
+            subpath = subpath.lstrip('/')
+            remoteRoot = posixpath.join(remoteDir, subpath)
             for f in files:
-                remoteRoot = remoteDir + '/' + parts[1]
-                if (remoteRoot.endswith('/')):
-                    remoteName = remoteRoot + f
-                else:
-                    remoteName = remoteRoot + '/' + f
+                remoteName = posixpath.join(remoteRoot, f)
 
-                if (parts[1] == ""):
+                if subpath == "":
                     remoteRoot = remoteDir
 
                 parent = os.path.dirname(remoteName)
@@ -388,8 +387,7 @@ class DeviceManagerSUT(DeviceManager):
                     self.mkDirs(remoteName)
                     existentDirectories.append(parent)
 
-                self.pushFile(os.path.join(root, f), remoteName, retryLimit=retryLimit)
-
+                self.pushFile(os.path.join(root, f), remoteName, retryLimit=retryLimit, createDir=False)
 
     def dirExists(self, remotePath):
         ret = self._runCmds([{ 'cmd': 'isdir ' + remotePath }]).strip()
@@ -401,25 +399,29 @@ class DeviceManagerSUT(DeviceManager):
     def fileExists(self, filepath):
         # Because we always have / style paths we make this a lot easier with some
         # assumptions
-        s = filepath.split('/')
-        containingpath = '/'.join(s[:-1])
-        return s[-1] in self.listFiles(containingpath)
+        filepath = posixpath.normpath(filepath)
+        # / should always exist but we can use this to check for things like
+        # having access to the filesystem
+        if filepath == '/':
+            return self.dirExists(filepath)
+        (containingpath, filename) = posixpath.split(filepath)
+        return filename in self.listFiles(containingpath)
 
     def listFiles(self, rootdir):
-        rootdir = rootdir.rstrip('/')
-        if (self.dirExists(rootdir) == False):
+        rootdir = posixpath.normpath(rootdir)
+        if not self.dirExists(rootdir):
             return []
         data = self._runCmds([{ 'cmd': 'cd ' + rootdir }, { 'cmd': 'ls' }])
 
         files = filter(lambda x: x, data.splitlines())
         if len(files) == 1 and files[0] == '<empty>':
-            # special case on the agent: empty directories return just the string "<empty>"
+            # special case on the agent: empty directories return just the
+            # string "<empty>"
             return []
         return files
 
     def removeFile(self, filename):
-        if (self.debug>= 2):
-            print "removing file: " + filename
+        self._logger.info("removing file: " + filename)
         if self.fileExists(filename):
             self._runCmds([{ 'cmd': 'rm ' + filename }])
 
@@ -427,6 +429,12 @@ class DeviceManagerSUT(DeviceManager):
         if self.dirExists(remoteDir):
             self._runCmds([{ 'cmd': 'rmdr ' + remoteDir }])
 
+    def moveTree(self, source, destination):
+        self._runCmds([{ 'cmd': 'mv %s %s' % (source, destination) }])
+
+    def copyTree(self, source, destination):
+        self._runCmds([{ 'cmd': 'dd if=%s of=%s' % (source, destination) }])
+
     def getProcessList(self):
         data = self._runCmds([{ 'cmd': 'ps' }])
 
@@ -444,8 +452,8 @@ class DeviceManagerSUT(DeviceManager):
                         # unexpected format
                         raise ValueError
                 except ValueError:
-                    print "ERROR: Unable to parse process list (bug 805969)"
-                    print "Line: %s\nFull output of process list:\n%s" % (line, data)
+                    self._logger.error("Unable to parse process list (bug 805969)")
+                    self._logger.error("Line: %s\nFull output of process list:\n%s" % (line, data))
                     raise DMError("Invalid process line: %s" % line)
 
         return processTuples
@@ -461,11 +469,10 @@ class DeviceManagerSUT(DeviceManager):
         if not appname:
             raise DMError("Automation Error: fireProcess called with no command to run")
 
-        if (self.debug >= 2):
-            print "FIRE PROC: '" + appname + "'"
+        self._logger.info("FIRE PROC: '%s'" % appname)
 
         if (self.processExist(appname) != None):
-            print "WARNING: process %s appears to be running already\n" % appname
+            self._logger.warn("process %s appears to be running already\n" % appname)
             if (failIfRunning):
                 raise DMError("Automation Error: Process is already running")
 
@@ -484,8 +491,7 @@ class DeviceManagerSUT(DeviceManager):
             time.sleep(1)
             waited += 1
 
-        if (self.debug >= 4):
-            print "got pid: %s for process: %s" % (pid, appname)
+        self._logger.debug("got pid: %s for process: %s" % (pid, appname))
         return pid
 
     def launchProcess(self, cmd, outputFile = "process.txt", cwd = '', env = '', failIfRunning=False):
@@ -500,10 +506,12 @@ class DeviceManagerSUT(DeviceManager):
         DEPRECATED: Use shell() or launchApplication() for new code
         """
         if not cmd:
-            if (self.debug >= 1):
-                print "WARNING: launchProcess called without command to run"
+            self._logger.warn("launchProcess called without command to run")
             return None
 
+        if cmd[0] == 'am' and hasattr(self, '_getExtraAmStartArgs'):
+            cmd = cmd[:2] + self._getExtraAmStartArgs() + cmd[2:] 
+
         cmdline = subprocess.list2cmdline(cmd)
         if (outputFile == "process.txt" or outputFile == None):
             outputFile = self.getDeviceRoot();
@@ -516,12 +524,20 @@ class DeviceManagerSUT(DeviceManager):
         cmdline = '%s %s' % (self._formatEnvString(env), cmdline)
 
         # fireProcess may trigger an exception, but we won't handle it
-        self.fireProcess(cmdline, failIfRunning)
+        if cmd[0] == "am":
+            # Robocop tests spawn "am instrument". sutAgent's exec ensures that
+            # am has started before returning, so there is no point in having
+            # fireProcess wait for it to start. Also, since "am" does not show
+            # up in the process list while the test is running, waiting for it
+            # in fireProcess is difficult.
+            self.fireProcess(cmdline, failIfRunning, 0)
+        else:
+            self.fireProcess(cmdline, failIfRunning)
         return outputFile
 
-    def killProcess(self, appname, forceKill=False):
-        if forceKill:
-            print "WARNING: killProcess(): forceKill parameter unsupported on SUT"
+    def killProcess(self, appname, sig=None):
+        if sig:
+            self._logger.warn("killProcess(): sig parameter unsupported on SUT")
         retries = 0
         while retries < self.retryLimit:
             try:
@@ -530,17 +546,16 @@ class DeviceManagerSUT(DeviceManager):
                 return
             except DMError, err:
                 retries +=1
-                print ("WARNING: try %d of %d failed to kill %s" %
+                self._logger.warn("try %d of %d failed to kill %s" %
                        (retries, self.retryLimit, appname))
-                if self.debug >= 4:
-                    print err
+                self._logger.debug(err)
                 if retries >= self.retryLimit:
                     raise err
 
     def getTempDir(self):
         return self._runCmds([{ 'cmd': 'tmpd' }]).strip()
 
-    def pullFile(self, remoteFile):
+    def pullFile(self, remoteFile, offset=None, length=None):
         # The "pull" command is different from other commands in that DeviceManager
         # has to read a certain number of bytes instead of just reading to the
         # next prompt.  This is more robust than the "cat" command, which will be
@@ -549,7 +564,7 @@ class DeviceManagerSUT(DeviceManager):
 
         def err(error_msg):
             err_str = 'DeviceManager: pull unsuccessful: %s' % error_msg
-            print err_str
+            self._logger.error(err_str)
             self._sock = None
             raise DMError(err_str)
 
@@ -557,30 +572,23 @@ class DeviceManagerSUT(DeviceManager):
         # the class level if we wanted to refactor sendCMD().  For now they are
         # only used to pull files.
 
-        def uread(to_recv, error_msg, timeout=None):
+        def uread(to_recv, error_msg):
             """ unbuffered read """
-            timer = 0
-            select_timeout = 1
-            if not timeout:
-                timeout = self.default_timeout
-
             try:
-                if select.select([self._sock], [], [], select_timeout)[0]:
+                data = ""
+                if select.select([self._sock], [], [], self.default_timeout)[0]:
                     data = self._sock.recv(to_recv)
-                    timer = 0
-                timer += select_timeout
-                if timer > timeout:
-                    err('timeout in uread while retrieving file')
-
                 if not data:
+                    # timed out waiting for response or error response
                     err(error_msg)
+
                 return data
             except:
                 err(error_msg)
 
         def read_until_char(c, buf, error_msg):
             """ read until 'c' is found; buffer rest """
-            while not '\n' in buf:
+            while not c in buf:
                 data = uread(1024, error_msg)
                 buf += data
             return buf.partition(c)
@@ -602,14 +610,20 @@ class DeviceManagerSUT(DeviceManager):
         # <filename>,-1\n<error message>
 
         # just send the command first, we read the response inline below
-        self._runCmds([{ 'cmd': 'pull ' + remoteFile }])
+        if offset is not None and length is not None:
+            cmd = 'pull %s %d %d' % (remoteFile, offset, length)
+        elif offset is not None:
+            cmd = 'pull %s %d' % (remoteFile, offset)
+        else: 
+            cmd = 'pull %s' % remoteFile
+
+        self._runCmds([{ 'cmd': cmd }])
 
         # read metadata; buffer the rest
         metadata, sep, buf = read_until_char('\n', buf, 'could not find metadata')
         if not metadata:
             return None
-        if self.debug >= 3:
-            print 'metadata: %s' % metadata
+        self._logger.debug('metadata: %s' % metadata)
 
         filename, sep, filesizestr = metadata.partition(',')
         if sep == '':
@@ -648,15 +662,13 @@ class DeviceManagerSUT(DeviceManager):
                           remoteFile)
 
     def getDirectory(self, remoteDir, localDir, checkDir=True):
-        if (self.debug >= 2):
-            print "getting files in '" + remoteDir + "'"
+        self._logger.info("getting files in '%s'" % remoteDir)
         if checkDir and not self.dirExists(remoteDir):
             raise DMError("Automation Error: Error getting directory: %s not a directory" %
                           remoteDir)
 
         filelist = self.listFiles(remoteDir)
-        if (self.debug >= 3):
-            print filelist
+        self._logger.debug(filelist)
         if not os.path.exists(localDir):
             os.makedirs(localDir)
 
@@ -684,8 +696,7 @@ class DeviceManagerSUT(DeviceManager):
 
     def _getRemoteHash(self, filename):
         data = self._runCmds([{ 'cmd': 'hash ' + filename }]).strip()
-        if self.debug >= 3:
-            print "remote hash returned: '%s'" % data
+        self._logger.debug("remote hash returned: '%s'" % data)
         return data
 
     def getDeviceRoot(self):
@@ -722,64 +733,78 @@ class DeviceManagerSUT(DeviceManager):
 
         self._runCmds([{ 'cmd': 'unzp %s %s' % (filePath, destDir)}])
 
-    def _wait_for_reboot(self, host, port):
-        if self.debug >= 3:
-            print 'Creating server with %s:%d' % (host, port)
-        timeout_expires = time.time() + self.reboot_timeout
+    def _getRebootServerSocket(self, ipAddr):
+        serverSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        serverSocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
+        serverSocket.settimeout(60.0)
+        serverSocket.bind((ipAddr, 0))
+        serverSocket.listen(1)
+        self._logger.debug('Created reboot callback server at %s:%d' %
+                           serverSocket.getsockname())
+        return serverSocket
+
+    def _waitForRebootPing(self, serverSocket):
         conn = None
-        data = ''
-        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
-        s.settimeout(60.0)
-        s.bind((host, port))
-        s.listen(1)
-        while not data and time.time() < timeout_expires:
+        data = None
+        startTime = datetime.datetime.now()
+        waitTime = datetime.timedelta(seconds=self.reboot_timeout)
+        while not data and datetime.datetime.now() - startTime < waitTime:
+            self._logger.info("Waiting for reboot callback ping from device...")
             try:
                 if not conn:
-                    conn, _ = s.accept()
+                    conn, _ = serverSocket.accept()
                 # Receiving any data is good enough.
                 data = conn.recv(1024)
                 if data:
+                    self._logger.info("Received reboot callback ping from device!")
                     conn.sendall('OK')
                 conn.close()
             except socket.timeout:
-                print '.'
+                pass
             except socket.error, e:
                 if e.errno != errno.EAGAIN and e.errno != errno.EWOULDBLOCK:
                     raise
-        if data:
-            # Sleep to ensure not only we are online, but all our services are
-            # also up.
-            time.sleep(self.reboot_settling_time)
-        else:
-            print 'Automation Error: Timed out waiting for reboot callback.'
-        s.close()
-        return data
 
-    def reboot(self, ipAddr=None, port=30000):
+        if not data:
+            raise DMError('Timed out waiting for reboot callback.')
+
+        self._logger.info("Sleeping for %s seconds to wait for device "
+                          "to 'settle'" % self.reboot_settling_time)
+        time.sleep(self.reboot_settling_time)
+
+
+    def reboot(self, ipAddr=None, port=30000, wait=False):
+        # port ^^^ is here for backwards compatibility only, we now
+        # determine a port automatically and safely
+        wait = (wait or ipAddr)
+
         cmd = 'rebt'
 
-        if self.debug > 3:
-            print "INFO: sending rebt command"
+        self._logger.info("Rebooting device")
 
-        if ipAddr is not None:
+        # if we're waiting, create a listening server and pass information on
+        # it to the device before rebooting (we do this instead of just polling
+        # to make sure the device actually rebooted -- yes, there are probably
+        # simpler ways of doing this like polling uptime, but this is what we're
+        # doing for now)
+        if wait:
+            if not ipAddr:
+                ipAddr = moznetwork.get_ip()
+            serverSocket = self._getRebootServerSocket(ipAddr)
             # The update.info command tells the SUTAgent to send a TCP message
             # after restarting.
             destname = '/data/data/com.mozilla.SUTAgentAndroid/files/update.info'
-            data = "%s,%s\rrebooting\r" % (ipAddr, port)
+            data = "%s,%s\rrebooting\r" % serverSocket.getsockname()
             self._runCmds([{'cmd': 'push %s %s' % (destname, len(data)),
                             'data': data}])
+            cmd += " %s %s" % serverSocket.getsockname()
 
-            ip, port = self._getCallbackIpAndPort(ipAddr, port)
-            cmd += " %s %s" % (ip, port)
-
-        status = self._runCmds([{'cmd': cmd}])
-
-        if ipAddr is not None:
-            status = self._wait_for_reboot(ipAddr, port)
-
-        if self.debug > 3:
-            print "INFO: rebt- got status back: " + str(status)
+        # actually reboot device
+        self._runCmds([{'cmd': cmd}])
+        # if we're waiting, wait for a callback ping from the agent before
+        # continuing (and throw an exception if we don't receive said ping)
+        if wait:
+            self._waitForRebootPing(serverSocket)
 
     def getInfo(self, directive=None):
         data = None
@@ -810,8 +835,7 @@ class DeviceManagerSUT(DeviceManager):
                     proclist.append(l.split('\t'))
             result['process'] = proclist
 
-        if (self.debug >= 3):
-            print "results: " + str(result)
+        self._logger.debug("results: %s" % result)
         return result
 
     def installApp(self, appBundlePath, destPath=None):
@@ -821,10 +845,8 @@ class DeviceManagerSUT(DeviceManager):
 
         data = self._runCmds([{ 'cmd': cmd }])
 
-        f = re.compile('Failure')
-        for line in data.split():
-            if (f.match(line)):
-                raise DMError("Remove Device Error: Error installing app. Error message: %s" % data)
+        if 'installation complete [0]' not in data:
+            raise DMError("Remove Device Error: Error installing app. Error message: %s" % data)
 
     def uninstallApp(self, appName, installPath=None):
         cmd = 'uninstall ' + appName
@@ -833,8 +855,7 @@ class DeviceManagerSUT(DeviceManager):
         data = self._runCmds([{ 'cmd': cmd }])
 
         status = data.split('\n')[0].strip()
-        if self.debug > 3:
-            print "uninstallApp: '%s'" % status
+        self._logger.debug("uninstallApp: '%s'" % status)
         if status == 'Success':
             return
         raise DMError("Remote Device Error: uninstall failed for %s" % appName)
@@ -845,12 +866,15 @@ class DeviceManagerSUT(DeviceManager):
             cmd += ' ' + installPath
         data = self._runCmds([{ 'cmd': cmd }])
 
-        if (self.debug > 3):
-            print "uninstallAppAndReboot: " + str(data)
+        self._logger.debug("uninstallAppAndReboot: %s" % data)
         return
 
-    def updateApp(self, appBundlePath, processName=None, destPath=None, ipAddr=None, port=30000):
-        status = None
+    def updateApp(self, appBundlePath, processName=None, destPath=None,
+                  ipAddr=None, port=30000, wait=False):
+        # port ^^^ is here for backwards compatibility only, we now
+        # determine a port automatically and safely
+        wait = (wait or ipAddr)
+
         cmd = 'updt '
         if processName is None:
             # Then we pass '' for processName
@@ -861,40 +885,21 @@ class DeviceManagerSUT(DeviceManager):
         if destPath:
             cmd += " " + destPath
 
-        if ipAddr is not None:
-            ip, port = self._getCallbackIpAndPort(ipAddr, port)
-            cmd += " %s %s" % (ip, port)
+        if wait:
+            if not ipAddr:
+                ipAddr = moznetwork.get_ip()
+            serverSocket = self._getRebootServerSocket(ipAddr)
+            cmd += " %s %s" % serverSocket.getsockname()
 
-        if self.debug >= 3:
-            print "INFO: updateApp using command: " + str(cmd)
+        self._logger.debug("updateApp using command: " % cmd)
 
-        status = self._runCmds([{'cmd': cmd}])
+        self._runCmds([{'cmd': cmd}])
 
-        if ipAddr is not None:
-            status = self._wait_for_reboot(ip, port)
-
-        if self.debug >= 3:
-            print "INFO: updateApp: got status back: %s" + str(status)
+        if wait:
+            self._waitForRebootPing(serverSocket)
 
     def getCurrentTime(self):
-        return self._runCmds([{ 'cmd': 'clok' }]).strip()
-
-    def _getCallbackIpAndPort(self, aIp, aPort):
-        """
-        Connect the ipaddress and port for a callback ping.
-
-        Defaults to current IP address and ports starting at 30000.
-        NOTE: the detection for current IP address only works on Linux!
-        """
-        ip = aIp
-        nettools = NetworkTools()
-        if (ip == None):
-            ip = nettools.getLanIp()
-        if (aPort != None):
-            port = nettools.findOpenPort(ip, aPort)
-        else:
-            port = nettools.findOpenPort(ip, 30000)
-        return ip, port
+        return int(self._runCmds([{ 'cmd': 'clok' }]).strip())
 
     def _formatEnvString(self, env):
         """
@@ -922,14 +927,12 @@ class DeviceManagerSUT(DeviceManager):
         supported resolutions: 640x480, 800x600, 1024x768, 1152x864, 1200x1024, 1440x900, 1680x1050, 1920x1080
         """
         if self.getInfo('os')['os'][0].split()[0] != 'harmony-eng':
-            if (self.debug >= 2):
-                print "WARNING: unable to adjust screen resolution on non Tegra device"
+            self._logger.warn("unable to adjust screen resolution on non Tegra device")
             return False
 
         results = self.getInfo('screen')
         parts = results['screen'][0].split(':')
-        if (self.debug >= 3):
-            print "INFO: we have a current resolution of %s, %s" % (parts[1].split()[0], parts[2].split()[0])
+        self._logger.debug("we have a current resolution of %s, %s" % (parts[1].split()[0], parts[2].split()[0]))
 
         #verify screen type is valid, and set it to the proper value (https://bugzilla.mozilla.org/show_bug.cgi?id=632895#c4)
         screentype = -1
@@ -950,8 +953,7 @@ class DeviceManagerSUT(DeviceManager):
         if (height < 100 or height > 9999):
             return False
 
-        if (self.debug >= 3):
-            print "INFO: adjusting screen resolution to %s, %s and rebooting" % (width, height)
+        self._logger.debug("adjusting screen resolution to %s, %s and rebooting" % (width, height))
 
         self._runCmds([{ 'cmd': "exec setprop persist.tegra.dpy%s.mode.width %s" % (screentype, width) }])
         self._runCmds([{ 'cmd': "exec setprop persist.tegra.dpy%s.mode.height %s" % (screentype, height) }])
