@@ -4,8 +4,9 @@
 
 "use strict";
 
-const {Cc, Ci} = require("chrome");
-const promise = require("sdk/core/promise");
+const {Cc, Ci, Cu} = require("chrome");
+const Services = require("Services");
+const {Promise: promise} = Cu.import("resource://gre/modules/Promise.jsm", {});
 const protocol = require("devtools/server/protocol");
 const {Arg, Option, method, RetVal, types} = protocol;
 const events = require("sdk/event/core");
@@ -13,7 +14,6 @@ const object = require("sdk/util/object");
 const { Class } = require("sdk/core/heritage");
 const { StyleSheetActor } = require("devtools/server/actors/stylesheets");
 
-loader.lazyImporter(this, "Services", "resource://gre/modules/Services.jsm");
 loader.lazyGetter(this, "CssLogic", () => require("devtools/styleinspector/css-logic").CssLogic);
 loader.lazyGetter(this, "DOMUtils", () => Cc["@mozilla.org/inspector/dom-utils;1"].getService(Ci.inIDOMUtils));
 
@@ -357,6 +357,10 @@ var PageStyleActor = protocol.ActorClass({
    */
   addElementRules: function(element, inherited, options, rules)
   {
+    if (!element.style) {
+      return;
+    }
+
     let elementStyle = this._styleRef(element);
 
     if (!inherited || this._hasInheritedProps(element.style)) {
@@ -402,7 +406,8 @@ var PageStyleActor = protocol.ActorClass({
         rules.push({
           rule: ruleActor,
           inherited: inherited,
-          pseudoElement: pseudo
+          pseudoElement: pseudo,
+          isSystem: isSystem
         });
       }
 
